// This software including the source code and documentation (the "SOFTWARE")
// is copyrighted.  You are free to use the SOFTWARE for any non-commercial
// purpose whatsoever provided you do not remove this copyright notice.
// If you wish to use the SOFTWARE for commercial purposes, please contact us
// to obtain a license.  The SOFTWARE is provided "as is", with absolutely no
// warranty.  Use it at your own risk.
// Website: http://www.math.ie/checker.html

#ifndef _myset_H
#define _myset_H  1

#include "unique.h"
#include <iostream>
using namespace std;

// convert a coordinate pair (i,j) to 0-based index (0-80)
//#define Coords2Idx( a ) ((a) % 10) - 1 + (((a) / 10) - 1)*9
// now table-based implementation; table is in misc.cpp
extern int g_c2i[ 100 ];
#define Coords2Idx(a) g_c2i[a]

// to store a set of coordinates (= pair of integers)
class MySet
   {
   public:
      MySet( bool swapped ) : _count( 0 ), _swapped( swapped ) {}
      
      void AddElement( int i, int j )
         {
         if ( _swapped )
            _set[ _count ] = (j + 1) * 10 + (i + 1);
         else
            _set[ _count ] = (i + 1) * 10 + (j + 1);
         _count++;
         }

      int GetSize() { return _count; }

      // does this set hit the diagonal or backdiagonal?
      bool DoesHitDiagonals() const
         {
         for ( int i = 0;  i < _count;  i++ )
            {
            int iRow = _set[ i ] / 10;
            int iCol = _set[ i ] % 10;
            if ( iRow == iCol  ||  iRow + iCol == 10 )
               return true;
            }
         return false;
         }

      // tests if this set is actually unavoidable by deleting all clues 
      // in the set from the completed grid and then seeing if what remains
      // has a multiple completions (useful for Sudoku-X)
      bool IsUnavoidable( char const* pszGrid ) const
         {
         char szPuzzle[ 82 ];
         memcpy( szPuzzle, pszGrid, 81 );
         szPuzzle[ 81 ] = '\0';
         for ( int i = 0;  i < _count;  i++ )
            szPuzzle[ Coords2Idx( _set[ i ] ) ] = '0';
         return ( GetNumberOfCompletionsOfPuzzle( szPuzzle ) > 1 );
         }

      void Sort()
         {
         // bubble sort is fine for small array
         for ( int i = 0;  i < _count - 1;  i++ )
            for ( int j = 0;  j < _count - 1 - i;  j++ )
               if ( _set[ j ] > _set[ j + 1 ] )
                  {
                  // swap
                  int tmp       = _set[ j ];
                  _set[ j ]     = _set[ j + 1 ];
                  _set[ j + 1 ] = tmp;
                  }
         }

      void Print( ostream& out = std::cout ) const
         {
         out << "{";
         for ( int i = 0;  i < _count;  i++ )
            {
            out << _set[ i ];
            if ( i != _count - 1 )
               out << ",";
            }
         out << "}" << endl;
         }

      bool Disjoint( MySet const& rhs )
         {
         for ( int i = 0;  i < _count;  i++ )
            for ( int j = 0;  j < rhs._count;  j++ )
               if ( _set[ i ] == rhs._set[ j ] )
                  return false;

         return true;
         }

      bool operator<( MySet const& rhs ) const
         {
         int diff = _count - rhs._count;
         if ( diff == 0 )
            {
            // both sets have the same size, need to look at elements
            for ( int i = 0;  i < _count  &&  diff == 0;  i++ )
               diff = _set[ i ] - rhs._set[ i ];
            }
         return (diff < 0);
         }

      bool operator==( MySet const& rhs ) const
         {
         if ( _count != rhs._count )
            return false;
         for ( int i = 0;  i < _count;  i++ )
            if ( _set[ i ] != rhs._set[ i ] )
               return false;
         return true;
         }

      int operator[]( int i ) const
         {
         return _set[ i ];
         }

      friend class UnavSet;

   private:
      int   _set[ 81 ];
      int   _count;
      bool  _swapped; // swap coordinates yes/no?
   };

#endif
